﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace ColourWheelVisualiser
{
	public partial class MainForm : Form
	{
		public MainForm()
		{
			InitializeComponent();
			comboBoxOutputType.SelectedIndex = 0;
		}


		public struct Colour
		{
			public Colour(float r, float g, float b)
			{
				this.r = r;
				this.g = g;
				this.b = b;
			}

			public float r, g, b;
		}


		public struct HSV
		{
			public HSV(float h, float s, float v)
			{
				this.h = h;
				this.s = s;
				this.v = v;
			}

			public float h, s, v;
		}


		private void selectImageButtonClick(object sender, EventArgs e)
		{
			OpenFileDialog openFileDialog = new OpenFileDialog();
			openFileDialog.InitialDirectory = "Environment.CurrentDirectory";
			openFileDialog.Filter = "image files|*.jpg;*.png";
			openFileDialog.FilterIndex = 0;
			openFileDialog.RestoreDirectory = true;

			if (openFileDialog.ShowDialog() == DialogResult.OK)
			{
				Stream imageStream = null;
				if ((imageStream = openFileDialog.OpenFile()) != null)
				{
					fileName = openFileDialog.FileName;

					using (imageStream)
					{
						//Read and parse the image
						BitmapFrame bitmapFrame;
						if (fileName.IndexOf(".png", StringComparison.OrdinalIgnoreCase) >= 0)
						{
							PngBitmapDecoder decoder = new PngBitmapDecoder(imageStream, BitmapCreateOptions.PreservePixelFormat, BitmapCacheOption.Default);
							bitmapFrame = decoder.Frames[0];
						}
						else
						{
							JpegBitmapDecoder decoder = new JpegBitmapDecoder(imageStream, BitmapCreateOptions.PreservePixelFormat, BitmapCacheOption.Default);
							bitmapFrame = decoder.Frames[0];
						}	

						//Get the pixels from the image						
						int stride = bitmapFrame.PixelWidth * inputBytesPerPixel;
						int size = bitmapFrame.PixelHeight * stride;
						pixels = new byte[size];
						bitmapFrame.CopyPixels(pixels, stride, 0);
						imageWidth = bitmapFrame.PixelWidth;
						imageHeight = bitmapFrame.PixelHeight;
					}

					reprocessImage();
				}
			}
		}


		public static HSV rgbToHsv(Colour rgb)
		{
			//Convert to HSV (these are standard formulas for this conversion)
			float cMax = max(rgb.r, rgb.g, rgb.b);
			float cMin = min(rgb.r, rgb.g, rgb.b);
			float diff = cMax - cMin;
			
			HSV result = new HSV();
			
			if (diff <= 0)							   result.h = 0;
			else if (rgb.r >= rgb.g && rgb.r >= rgb.b) result.h = 60 * (((rgb.g - rgb.b) / diff) % 6);
			else if (rgb.g >= rgb.r && rgb.g >= rgb.b) result.h = 60 * (((rgb.b - rgb.r) / diff) + 2);
			else									   result.h = 60 * (((rgb.r - rgb.g) / diff) + 4);
			
			//Wrap around to make sure we have a positive number here
			if (result.h < 0)
			{
				result.h += 360;
			}
									
			if (cMax <= 0)	result.s = 0;
			else			result.s = diff / cMax;

			result.v = cMax;

			return result;
		}


		public static Colour hsvToRgb(HSV hsv)
		{
			float c = hsv.v * hsv.s;
			float x = c * (1 - Math.Abs((hsv.h / 60) % 2 - 1));
			float m = hsv.v - c;

			Colour result;
			if		(hsv.h < 60)	result = new Colour(c, x, 0);
			else if (hsv.h < 120)	result = new Colour(x, c, 0);
			else if (hsv.h < 180)	result = new Colour(0, c, x);
			else if (hsv.h < 240)	result = new Colour(0, x, c);
			else if (hsv.h < 300)	result = new Colour(x, 0, c);
			else					result = new Colour(c, 0, x);
			result.r += m;
			result.g += m;
			result.b += m;
			return result;
		}


		public static float max(float a, float b, float c)
		{
			if		(a >= b && a >= c)	return a;
			else if (b >= c && b >= a)	return b;
			else						return c;
		}


		public static float min(float a, float b, float c)
		{
			if (a <= b && a <= c) return a;
			else if (b <= c && b <= a) return b;
			else return c;
		}


		private void reprocessImage()
		{
			//Get the divider setting that has been set and correct if set incorrectly
			float standardDivider;
			bool parsedSuccessfully = float.TryParse(textBoxDivider.Text, out standardDivider);
			if (!parsedSuccessfully)
			{
				standardDivider = 10;
			}

			//Get the output image resolution
			int outputImageDimensions;
			parsedSuccessfully = int.TryParse(resolutionTextBox.Text, out outputImageDimensions);
			if (!parsedSuccessfully)
			{
				outputImageDimensions = 200;
			}

			bool outputAsColourWheel = comboBoxOutputType.SelectedIndex == 0;

			if (pixels != null)
			{
				//Create a new buffer to write colours to
				int outputBytesPerPixel = 3;
				int numOutputValues = outputImageDimensions * outputImageDimensions * outputBytesPerPixel;
				float[] outputPixels = new float[numOutputValues];

				//Initialise the buffer to black
				for (int x = 0; x < outputImageDimensions; x++)
				{
					for (int y = 0; y < outputImageDimensions; y++)
					{
						int pixelIndex = (x * outputImageDimensions + y) * outputBytesPerPixel;
						outputPixels[pixelIndex + 0] = 0; //red
						outputPixels[pixelIndex + 1] = 0; //green
						outputPixels[pixelIndex + 2] = 0; //blue
					}
				}

				//Process each individual pixel
				for (int y = 0; y < imageHeight; y++)
				{
					for (int x = 0; x < imageWidth; x++)
					{
						//Get pixel colour
						int index = (y * imageWidth + x) * inputBytesPerPixel;
						Colour colour = new Colour(pixels[index + 2] / 255.0f,
													pixels[index + 1] / 255.0f,
													pixels[index + 0] / 255.0f);

						//Convert to hue/saturation/value
						HSV hsv = rgbToHsv(colour);

						//Convert back to RGB but at full brightness
						Colour colourWithNeutralValue = hsvToRgb(new HSV(hsv.h, hsv.s, 1));

						int xPos, yPos;
						if (outputAsColourWheel)
						{
							//Show as HSV circle
							float value = (hsv.h + 150) / 360 * 2 * (float)Math.PI;
							float xNormalised = (float)Math.Sin(value);
							float yNormalised = (float)Math.Cos(value);

							//Make the circle a little bit smaller to clearly show pixels exactly at the rim of the circle
							//(otherwise those would be difficult to see due to the white area outside the circle)
							xNormalised *= 0.97f;
							yNormalised *= 0.97f;

							xPos = (int)(outputImageDimensions / 2 + xNormalised * (outputImageDimensions / 2) * hsv.s);
							yPos = (int)(outputImageDimensions / 2 + yNormalised * (outputImageDimensions / 2) * hsv.s);
						}
						else
						{
							//Show as square HSV image
							xPos = (int)(hsv.h * outputImageDimensions / 360);
							yPos = (int)(hsv.s * outputImageDimensions);
						}

						xPos = Math.Min(xPos, outputImageDimensions - 1);
						yPos = Math.Min(yPos, outputImageDimensions - 1);
						int colourIndex = xPos * outputBytesPerPixel + yPos * outputImageDimensions * outputBytesPerPixel;
						outputPixels[colourIndex + 0] += colourWithNeutralValue.r;
						outputPixels[colourIndex + 1] += colourWithNeutralValue.g;
						outputPixels[colourIndex + 2] += colourWithNeutralValue.b;
					}
				}

				//Convert the floating point buffer to a byte buffer for the image
				byte[] outputPixelsByte = new byte[numOutputValues];
				for (int x = 0; x < outputImageDimensions; x++)
				{
					for (int y = 0; y < outputImageDimensions; y++)
					{
						int pixelIndex = (y * outputImageDimensions + x) * outputBytesPerPixel;

						float divider = standardDivider;
						divider = Math.Max(divider, outputPixels[pixelIndex + 0]);
						divider = Math.Max(divider, outputPixels[pixelIndex + 1]);
						divider = Math.Max(divider, outputPixels[pixelIndex + 2]);
						float multiplier = 1.0f / divider;

						if (outputAsColourWheel)
						{
							//Darken the colour towards the centre so that we have a grey centre instead of a white centre
							Vector offsetFromCentre = new Vector((float)(x) / outputImageDimensions * 2 - 1,
																	(float)(y) / outputImageDimensions * 2 - 1);
							float distanceFromCentre = (float)(offsetFromCentre.Length);
							distanceFromCentre = Math.Min(distanceFromCentre, 1.0f);
							multiplier *= distanceFromCentre * 0.5f + 0.5f;
						}
						else
						{
							//Darken the colour towards the top so that desaturated colour are grey instead of white
							multiplier *= ((float)(y) / outputImageDimensions) * 0.5f + 0.5f;
						}

						//Output a single colour outside the colour wheel if this is a colour wheel
						bool outputSingleColour = false;
						byte overwritingColour = 255;
						if (outputAsColourWheel)
						{
							Vector position = new Vector((float)(x) / outputImageDimensions * 2 - 1,
														 (float)(y) / outputImageDimensions * 2 - 1);
							double lengthSquared = position.LengthSquared;

							//Make the area outside the circle white, and also make the exact centre white, so that it is clear if a gamut is off-centre
							if (lengthSquared > 1 || lengthSquared < 0.0002)
							{
								outputSingleColour = true;
								overwritingColour = 255;
							}
						}

						if (outputSingleColour)
						{
							outputPixelsByte[pixelIndex + 0] = overwritingColour;
							outputPixelsByte[pixelIndex + 1] = overwritingColour;
							outputPixelsByte[pixelIndex + 2] = overwritingColour;
						}
						else
						{
							outputPixelsByte[pixelIndex + 0] = (byte)(outputPixels[pixelIndex + 0] * 255 * multiplier);
							outputPixelsByte[pixelIndex + 1] = (byte)(outputPixels[pixelIndex + 1] * 255 * multiplier);
							outputPixelsByte[pixelIndex + 2] = (byte)(outputPixels[pixelIndex + 2] * 255 * multiplier);
						}
					}
				}

				//Create an output bitmapsource to write to a file
				outputBitmapSource = BitmapSource.Create(outputImageDimensions, outputImageDimensions, //resolution
																	  100, 100,										//DPI
																	  PixelFormats.Rgb24,							//pixel format
																	  null,											//palette,
																	  outputPixelsByte,								//pixels array
																	  outputImageDimensions * outputBytesPerPixel);	//stride

				//Also show the result in the interface
				System.Drawing.Bitmap outputBitmap;
				using (MemoryStream outStream = new MemoryStream())
				{
					BitmapEncoder enc = new BmpBitmapEncoder();
					enc.Frames.Add(BitmapFrame.Create(outputBitmapSource));
					enc.Save(outStream);
					outputBitmap = new System.Drawing.Bitmap(outStream);
				}
				pictureBoxOutput.Image = outputBitmap;

				//Also show the original image in the interface
				pictureBoxInput.Image = Image.FromFile(fileName);
			}
		}


		private void saveToFileButtonClick(object sender, EventArgs e)
		{
			FileStream outputStream = new FileStream(fileName + "_gamut.bmp", FileMode.Create);
			BmpBitmapEncoder encoder = new BmpBitmapEncoder();
			encoder.Frames.Add(BitmapFrame.Create(outputBitmapSource));
			encoder.Save(outputStream);
			outputStream.Close();
		}


		private void somethingChanged(object sender, EventArgs e)
		{
		}


		private string fileName;
		private byte[] pixels;
		private int imageWidth, imageHeight;
		private int inputBytesPerPixel = 3;
		private BitmapSource outputBitmapSource;

		private void reprocessButton_Click(object sender, EventArgs e)
		{
			reprocessImage();
		}
	}
}